@@ -3,7 +3,8 @@ use crate::{
     util::{apply_unary_ops, CompositionOfUnaryOps},
     ExParseError,
 };
-use smallvec::{SmallVec, smallvec};
+use itertools::Itertools;
+use smallvec::{smallvec, SmallVec};
 use std::fmt::Debug;
 
 type ExprIdxVec = SmallVec<[usize; 32]>;
@@ -11,7 +12,7 @@ type ExprIdxVec = SmallVec<[usize; 32]>;
 /// Container of binary operators of one expression.
 pub type BinOpVec<T> = SmallVec<[BinOp<T>; 32]>;
 
-const N_NODES_ON_STACK: usize = 32usize;
+pub const N_NODES_ON_STACK: usize = 32usize;
 
 pub type FlatNodeVec<T> = SmallVec<[FlatNode<T>; N_NODES_ON_STACK]>;
 pub type FlatOpVec<T> = SmallVec<[FlatOp<T>; N_NODES_ON_STACK]>;
@@ -72,22 +73,24 @@ impl<T: Copy> FlatNode<T> {
 /// use exmex::{parse_with_default_ops};
 ///
 /// // create an expression by parsing a string
-/// let expr_parsed = parse_with_default_ops::<f32>("sin(1+{x})*{y}")?;
-/// let result_parsed = expr_parsed.eval(&[2.0, 1.5]);
-/// assert!((result_parsed - (1.0 + 2.0 as f32).sin() * 1.5).abs() < 1e-6);
+/// let expr = parse_with_default_ops::<f32>("sin(1+y)*x")?;
+/// assert!((expr.eval(&[2.0, 1.5])? - (1.0 + 2.0 as f32).sin() * 1.5).abs() < 1e-6);
 /// #
 /// #     Ok(())
 /// # }
 /// ```
-/// The second argument `&[2.0, 1.5]` in the call of [`eval`](FlatEx::eval) specifies
-/// that we want to evaluate the expression for the varibale values `x=2.0` and `y=1.5`.
-/// Note that variables need to be within curly brackets in the string to-be-parsed.
+/// The second argument `&[2.0, 1.5]` in the call of [`eval`](FlatEx::eval) specifies the
+/// variable values in the order of their occurrence in the string.
+/// In this example, we want to evaluate the expression for the varibale values `y=2.0` and `x=1.5`.
+/// Variables in the string to-be-parsed are all substrings that are no numbers, no 
+/// operators, and no parentheses.
 ///
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub struct FlatEx<T: Copy> {
     nodes: FlatNodeVec<T>,
     ops: FlatOpVec<T>,
     prio_indices: ExprIdxVec,
+    n_unique_vars: usize,
 }
 
 fn apply_uop_if_some<T: Copy>(uop: &Option<CompositionOfUnaryOps<T>>, val: T) -> T {
@@ -107,13 +110,21 @@ impl<T: Copy> FlatEx<T> {
     ///            the n-th variable as given in the string that has been parsed to this expression.
     ///            Thereby, only the first occurrence of the variable in the string is relevant.
     ///
-    /// # Panics
+    /// # Errors
     ///
-    /// If more variables are existent than elements in the argument `vars`, we panic due to
-    /// index out of bounds. Vice versa, if more arguments are passed than variables existent,
-    /// the last variables are ignored.
+    /// If the number of variables in the parsed expression are different from the length of
+    /// the variable slice, we return an [`ExParseError`](ExParseError).
     ///
-    pub fn eval(&self, vars: &[T]) -> T {
+    pub fn eval(&self, vars: &[T]) -> Result<T, ExParseError> {
+        if self.n_unique_vars != vars.len() {
+            return Err(ExParseError {
+                msg: format!(
+                    "parsed expression contains {} vars but passed slice has {} elements",
+                    self.n_unique_vars,
+                    vars.len()
+                ),
+            });
+        }
         let mut numbers = self
             .nodes
             .iter()
@@ -127,6 +138,7 @@ impl<T: Copy> FlatEx<T> {
                 )
             })
             .collect::<SmallVec<[T; 32]>>();
+        
         let mut ignore: SmallVec<[bool; N_NODES_ON_STACK]> = smallvec![false; N_NODES_ON_STACK];
         for (i, &bin_op_idx) in self.prio_indices.iter().enumerate() {
             let num_idx = self.prio_indices[i];
@@ -134,7 +146,7 @@ impl<T: Copy> FlatEx<T> {
             while ignore[num_idx - shift_left] {
                 shift_left += 1usize;
             }
-            let mut shift_right= 1usize;
+            let mut shift_right = 1usize;
             while ignore[num_idx + shift_right] {
                 shift_right += 1usize;
             }
@@ -146,7 +158,7 @@ impl<T: Copy> FlatEx<T> {
             };
             ignore[num_idx + shift_right] = true;
         }
-        numbers[0]
+        Ok(numbers[0])
     }
 
     fn compile(&mut self) {
@@ -227,16 +239,22 @@ fn flatten_vecs<T: Copy>(expr: &Expression<T>, prio_offset: i32) -> (FlatNodeVec
     }
 
     if expr.unary_ops.len() > 0 {
-        if flat_nodes.len() > 1 {
+        if flat_ops.len() > 0 {
             // find the last binary operator with the lowest priority of this expression,
             // since this will be executed as the last one
             let low_prio_op = match flat_ops.iter_mut().rev().min_by_key(|op| op.bin_op.prio) {
                 None => panic!("cannot have more than one flat node but no binary ops"),
                 Some(x) => x,
             };
+            let mut new_uops = expr.unary_ops.clone();
+            let mut existing_uops = match low_prio_op.unary_op.clone() {
+                Some(uops) => uops,
+                None => CompositionOfUnaryOps::<T>::new(),
+            };
+            new_uops.append(&mut existing_uops);
             *low_prio_op = FlatOp {
                 bin_op: low_prio_op.bin_op,
-                unary_op: Some(expr.unary_ops.clone()),
+                unary_op: Some(new_uops),
             }
         } else {
             let mut new_op = expr.unary_ops.clone();
@@ -364,10 +382,17 @@ impl<T: Copy + Debug> Expression<T> {
     pub fn flatten(&self) -> FlatEx<T> {
         let (nodes, ops) = flatten_vecs(self, 0);
         let indices = prioritized_indices_flat(&ops, &nodes);
+        let n_unique_vars = nodes.iter().filter_map(|n| {
+            match n.kind {
+                FlatNodeKind::Var(idx) => Some(idx),
+                _ => None,
+            }            
+        }).unique().count();
         let mut flatex = FlatEx {
             nodes: nodes,
             ops: ops,
             prio_indices: indices,
+            n_unique_vars: n_unique_vars,
         };
         flatex.compile();
         flatex
@@ -381,7 +406,7 @@ mod test {
     #[test]
     fn test_compile() {
         let flat_ex = parse_with_default_ops::<f64>("1*sin(2-0.1)").unwrap();
-        assert_float_eq_f64(flat_ex.eval(&[]), 1.9f64.sin());
+        assert_float_eq_f64(flat_ex.eval(&[]).unwrap(), 1.9f64.sin());
         assert_eq!(flat_ex.nodes.len(), 1);
     }
 }
